"""Configurator class to use different configurators like SMAC."""

from __future__ import annotations
from pathlib import Path
import shutil
import random

from runrunner import Runner, Run

from sparkle.configurator.configurator import Configurator
from sparkle.configurator.implementations.smac2 import SMAC2Scenario
from sparkle.solver import Solver
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.instance import InstanceSet
from sparkle.types import SparkleObjective


class ParamILS(Configurator):
    """Class for ParamILS (Java) configurator."""

    configurator_path = Path(__file__).parent / "ParamILS"
    configurator_executable = configurator_path / "paramils"
    target_algorithm = "paramils_target_algorithm.py"
    configurator_target = configurator_path / target_algorithm

    version = "3.0.0"
    full_name = "Parameter Iterated Local Search"

    def __init__(self: ParamILS) -> None:
        """Returns the ParamILS (Java) configurator, V3.0.0."""
        return super().__init__(multi_objective_support=False)

    @property
    def name(self: ParamILS) -> str:
        """Returns the name of the configurator."""
        return ParamILS.__name__

    @staticmethod
    def scenario_class() -> ParamILSScenario:
        """Returns the ParamILS scenario class."""
        return ParamILSScenario

    @staticmethod
    def check_requirements(verbose: bool = False) -> bool:
        """Check that ParamILS is installed."""
        import warnings

        if shutil.which("java") is None:
            if verbose:
                warnings.warn(
                    "ParamILS requires Java 1.8.0_402, but Java is not installed. "
                    "Please ensure Java is installed."
                )
            return False
        if not ParamILS.configurator_executable.exists():
            if verbose:
                warnings.warn(
                    "ParamILS executable not found. Please ensure ParamILS is installed "
                    f"in the expected Path ({ParamILS.configurator_path})."
                )
            return False
        return True

    @staticmethod
    def download_requirements(
        paramils_zip_url: str = "https://github.com/ADA-research/Sparkle/raw/refs/heads/"
        "development/Resources/Configurators/ParamILS-v.3.0.0.zip",
    ) -> None:
        """Download ParamILS."""
        if ParamILS.configurator_executable.exists():
            return  # Already installed
        from urllib.request import urlopen
        import zipfile
        import io

        r = urlopen(paramils_zip_url, timeout=60)
        z = zipfile.ZipFile(io.BytesIO(r.read()))
        z.extractall(ParamILS.configurator_path)
        ParamILS.configurator_executable.chmod(0o755)

    def configure(
        self: ParamILS,
        scenario: ParamILSScenario,
        data_target: PerformanceDataFrame,
        validate_after: bool = True,
        sbatch_options: list[str] = [],
        slurm_prepend: str | list[str] | Path = None,
        num_parallel_jobs: int = None,
        base_dir: Path = None,
        run_on: Runner = Runner.SLURM,
    ) -> list[Run]:
        """Start configuration job.

        Args:
            scenario: ConfigurationScenario object
            data_target: PerformanceDataFrame where to store the found configurations
            validate_after: Whether the Validator will be called after the configuration
            sbatch_options: List of slurm batch options to use
            slurm_prepend: Slurm script to prepend to the sbatch
            num_parallel_jobs: The maximum number of jobs to run parallel.
            base_dir: The path where the sbatch scripts will be created for Slurm.
            run_on: On which platform to run the jobs. Default: Slurm.

        Returns:
            A RunRunner Run object.
        """
        if shutil.which("java") is None:
            raise RuntimeError(
                "ParamILS requires Java 1.8.0_402, but Java is not installed. "
                "Please ensure Java is installed and try again."
            )
        scenario.create_scenario()
        configuration_ids = scenario.configuration_ids
        # The maximum seed size for ParamILS is 999 999 999
        seeds = [random.randint(0, 10**9 - 1) for _ in range(scenario.number_of_runs)]
        output = [
            f"{(scenario.results_directory).absolute()}/"
            f"{scenario.name}_seed_{config_id}_paramils.txt"
            for config_id in configuration_ids
        ]
        # NOTE: Could add --rungroup $dirname to change the created directory name
        cmds = [
            f"python3 {Configurator.configurator_cli_path.absolute()} "
            f"{ParamILS.__name__} {output_file} {data_target.csv_filepath} "
            f"{scenario.scenario_file_path} {configuration_id} "
            f"{ParamILS.configurator_executable.absolute()} "
            f"--scenario-file {scenario.scenario_file_path} "
            f"--seed {seed} "
            for output_file, configuration_id, seed in zip(
                output, configuration_ids, seeds
            )
        ]
        if num_parallel_jobs is not None:
            num_parallel_jobs = max(num_parallel_jobs, len(cmds))
        return super().configure(
            configuration_commands=cmds,
            data_target=data_target,
            output=output,
            slurm_prepend=slurm_prepend,
            num_parallel_jobs=num_parallel_jobs,
            scenario=scenario,
            configuration_ids=configuration_ids,
            validate_after=validate_after,
            sbatch_options=sbatch_options,
            base_dir=base_dir,
            run_on=run_on,
        )

    @staticmethod
    def organise_output(
        output_source: Path,
        output_target: Path = None,
        scenario: ParamILSScenario = None,
        configuration_id: str = None,
    ) -> None | dict:
        """Retrieves configurations from SMAC files and places them in output."""
        # Extract from log file
        configuration = {"configuration_id": configuration_id}
        skipping = True
        for line in output_source.open().readlines():
            if skipping:
                if "[INFO ] Differences with initial configuration:" in line:
                    skipping = False
                continue
            if ":" not in line or "->" not in line:
                break
            variable = line.split(":")[0].strip()
            value = line.split("->")[1].strip()
            configuration[variable] = value
        return Configurator.save_configuration(
            scenario, configuration_id, configuration, output_target
        )

    def get_status_from_logs(self: ParamILS) -> None:
        """Method to scan the log files of the configurator for warnings."""
        return


class ParamILSScenario(SMAC2Scenario):
    """Class to handle ParamILS configuration scenarios."""

    def __init__(
        self: ParamILSScenario,
        solver: Solver,
        instance_set: InstanceSet,
        sparkle_objectives: list[SparkleObjective],
        number_of_runs: int,
        parent_directory: Path,
        solver_calls: int = None,
        max_iterations: int = None,
        solver_cutoff_time: int = None,
        cli_cores: int = None,
        use_cpu_time_in_tunertime: bool = None,
        feature_data: FeatureDataFrame | Path = None,
        tuner_timeout: int = None,
        focused_ils: bool = True,
        initial_configurations: int = None,
        min_runs: int = None,
        max_runs: int = None,
        random_restart: float = None,
        timestamp: str = None,
    ) -> None:
        """Initialize scenario paths and names.

        Args:
            solver: Solver that should be configured.
            instance_set: Instances object for the scenario.
            sparkle_objectives: SparkleObjectives used for each run of the configuration.
            parent_directory: Directory in which the scenario should be created.
            number_of_runs: The number of configurator runs to perform
                for configuring the solver.
            solver_calls: The number of times the solver is called for each
                configuration run
            max_iterations: The maximum number of iterations allowed for each
                configuration run. [iteration-limit, numIterations, numberOfIterations]
            solver_cutoff_time: The maximum number of seconds allowed for each
                Solver call.
            cli_cores: The maximum number of cores allowed for each
                configuration run.
            use_cpu_time_in_tunertime: Whether to use cpu_time in the tuner
                time limit.
            feature_data: The feature data for the instances in the scenario.
            tuner_timeout: The maximum number of seconds allowed for the tuner.
            focused_ils: Comparison approach of ParamILS.
                True for focused ILS, false for basic.
            initial_configurations: The number of initial configurations.
            min_runs: The minimum number of runs required for a single configuration.
            max_runs: The maximum number of runs allowed for a single configuration.
            random_restart: The probability to restart from a random configuration.
            timestamp: An optional timestamp for the directory name.
        """
        super().__init__(
            solver,
            instance_set,
            sparkle_objectives,
            number_of_runs,
            parent_directory,
            solver_calls,
            max_iterations,
            None,
            None,
            solver_cutoff_time,
            None,
            cli_cores,
            use_cpu_time_in_tunertime,
            feature_data,
            timestamp,
        )
        self.tuner_timeout = tuner_timeout
        self.multi_objective = len(sparkle_objectives) > 1  # Not using MO yet in Sparkle
        self.focused = focused_ils
        self.initial_configurations = initial_configurations
        self.min_runs = min_runs
        self.max_runs = max_runs
        self.random_restart = random_restart

    @property
    def configurator(self: ParamILSScenario) -> ParamILS:
        """Return the type of configurator the scenario belongs to."""
        return ParamILS

    def create_scenario_file(self: ParamILSScenario) -> Path:
        """Create a file with the configuration scenario."""
        from sparkle.tools.parameters import PCSConvention

        scenario_file = super().create_scenario_file(
            ParamILS.configurator_target, PCSConvention.ParamILS
        )
        with scenario_file.open("+a") as fout:
            fout.write("check-instances-exist = True\n")
            if self.focused is not None:
                approach = "FOCUSED" if self.focused else "BASIC"
                fout.write(f"approach = {approach}\n")
            if self.initial_configurations:
                fout.write(f"R = {self.initial_configurations}\n")
            if self.min_runs:
                fout.write(f"min-runs = {self.min_runs}\n")
            if self.max_runs:
                fout.write(f"max-runs = {self.max_runs}\n")
            if self.random_restart:
                fout.write(f"random-restart = {self.random_restart}\n")
            if self.tuner_timeout:
                fout.write(f"tuner-timeout = {self.tuner_timeout}\n")
        return scenario_file

    @staticmethod
    def from_file(scenario_file: Path) -> ParamILSScenario:
        """Reads scenario file and initalises ConfigurationScenario."""
        from sparkle.types import resolve_objective
        from sparkle.instance import Instance_Set

        config = {}
        with scenario_file.open() as file:
            import ast

            for line in file:
                key, value = line.strip().split(" = ")
                key = key.replace("-", "_")
                try:
                    config[key] = ast.literal_eval(value)
                except Exception:
                    config[key] = value

        _, solver_path, _, objective_str = config["algo"].split(" ")
        objective = resolve_objective(objective_str)
        solver = Solver(Path(solver_path.strip()))
        # Extract the instance set from the instance file
        instance_file_path = Path(config["instance_file"])
        instance_set_path = Path(instance_file_path.open().readline().strip()).parent
        instance_set = Instance_Set(Path(instance_set_path))

        del config["algo"]
        del config["run_obj"]
        del config["deterministic"]
        del config["paramfile"]
        del config["instance_file"]
        del config["test_instance_file"]
        del config["outdir"]
        del config["validation"]
        del config["check_instances_exist"]

        if "cutoffTime" in config:
            config["solver_cutoff_time"] = config.pop("cutoffTime")
        if "runcount-limit" in config:
            config["solver_calls"] = config.pop("runcount-limit")
        if "approach" in config:
            config["focused_ils"] = config.pop("approach") == "FOCUS"
        if "R" in config:
            config["initial_configurations"] = config.pop("R")
        if "runcount_limit" in config:
            config["solver_calls"] = config.pop("runcount_limit")
        results_folder = scenario_file.parent / "results"
        number_of_runs = len([p for p in results_folder.iterdir() if p.is_file()])
        timestamp = scenario_file.parent.name.split("_")[-1]
        return ParamILSScenario(
            solver,
            instance_set,
            [objective],
            number_of_runs,
            scenario_file.parent.parent,
            **config,
            timestamp=timestamp,
        )



"""Helper classes/method for LaTeX and bibTeX."""

import math

import numpy as np
import pandas as pd
import plotly
import plotly.express as px
import pylatex as pl
import kaleido

kaleido.get_chrome_sync()  # Ensure chrome is available for Kaleido


class AutoRef(pl.base_classes.CommandBase):
    """AutoRef command for PyLateX."""

    _latex_name = "autoref"
    packages = [pl.Package("hyperref")]


def comparison_plot(
    data_frame: pd.DataFrame, title: str = None
) -> plotly.graph_objects.Figure:
    """Creates a comparison plot from the given data frame.

    The first column is used for the x axis, the second for the y axis.

    Args:
        data_frame: The data frame with the data
        x_label: The label for the x axis
        y_label: The label for the y axis
        title: The title of the plot
        output_plot: The path where the plot should be written to

    Returns:
        The plot object
    """
    from scipy import stats

    # Determine if data is log scale, linregress tells us how linear the data is
    linregress = stats.linregress(
        data_frame[data_frame.columns[0]].to_numpy(),
        data_frame[data_frame.columns[1]].to_numpy(),
    )
    log_scale = not (linregress.rvalue > 0.65 and linregress.pvalue < 0.05)

    if log_scale and (data_frame < 0).any(axis=None):
        # Log scale cannot deal with negative and zero values, set to smallest non zero
        data_frame[data_frame < 0] = np.nextafter(0, 1)

    # Maximum value should come from the objective?
    min_value, max_value = data_frame.min(axis=None), data_frame.max(axis=None)
    # Slightly more than min/max for interpretability
    if log_scale:  # Take next step on log scale
        max_value = 10 ** math.ceil(math.log(max_value, 10))
        plot_range = (min_value, max_value)
    else:  # Take previous/next step on linear scale
        order_magnitude = math.ceil(math.log(max_value, 10))
        next_step_max = math.ceil(max_value / (10 ** (order_magnitude - 1))) * 10 ** (
            order_magnitude - 1
        )
        plot_range = (0, next_step_max) if min_value > 0 else (min_value, next_step_max)
    fig = px.scatter(
        data_frame=data_frame,
        x=data_frame.columns[0],
        y=data_frame.columns[1],
        range_x=plot_range,
        range_y=plot_range,
        title=title,
        log_x=log_scale,
        log_y=log_scale,
        width=1000,
        height=1000,
    )
    # Add dividing diagonal
    fig.add_shape(
        type="line",
        x0=0,
        y0=0,
        x1=max_value,
        y1=max_value,
        line=dict(color="grey", dash="dot", width=1),
    )
    # Add maximum lines
    fig.add_shape(
        type="line",
        opacity=0.7,
        x0=0,
        y0=max_value,
        x1=max_value,
        y1=max_value,
        line=dict(color="red", width=1.5, dash="longdash"),
    )
    fig.add_shape(
        type="line",
        opacity=0.7,
        x0=max_value,
        y0=0,
        x1=max_value,
        y1=max_value,
        line=dict(color="red", width=0.5, dash="longdash"),
    )
    fig.update_traces(marker=dict(color="RoyalBlue", symbol="x"))
    fig.update_layout(plot_bgcolor="white", autosize=False, width=1000, height=1000)
    minor = dict(ticks="inside", ticklen=6, showgrid=True) if log_scale else None
    # Tick every 10^(log(max)) / 10 for linear scale, log scale is resolved by plotly
    dtick = 1 if log_scale else 10 ** (math.ceil(math.log(max_value, 10)) - 1)
    fig.update_xaxes(
        mirror=True,
        tickmode="linear",
        ticks="outside",
        tick0=0,
        minor=minor,
        dtick=dtick,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
    )
    fig.update_yaxes(
        mirror=True,
        tickmode="linear",
        ticks="outside",
        tick0=0,
        minor=minor,
        dtick=dtick,
        showline=True,
        linecolor="black",
        gridcolor="lightgrey",
    )
    return fig



"""Class for Sparkle Objective and Performance."""

from __future__ import annotations
from enum import Enum
import typing
import numpy as np
from sparkle.types.status import SolverStatus


class UseTime(str, Enum):
    """Enum describing what type of time to use."""

    WALL_TIME = "WALL_TIME"
    CPU_TIME = "CPU_TIME"
    NO = "NO"

    @classmethod
    def _missing_(cls: UseTime, value: object) -> UseTime:
        """Return error use time."""
        return UseTime.NO


class SparkleObjective:
    """Objective for Sparkle specified by user."""

    name: str
    run_aggregator: typing.Callable
    instance_aggregator: typing.Callable
    solver_aggregator: typing.Callable
    minimise: bool
    post_process: typing.Callable
    use_time: UseTime
    metric: bool

    def __init__(
        self: SparkleObjective,
        name: str,
        run_aggregator: typing.Callable = np.mean,
        instance_aggregator: typing.Callable = np.mean,
        solver_aggregator: typing.Callable = None,
        minimise: bool = True,
        post_process: typing.Callable = None,
        use_time: UseTime = UseTime.NO,
        metric: bool = False,
    ) -> None:
        """Create sparkle objective from string."""
        self.name = name
        self.run_aggregator: typing.Callable = run_aggregator
        self.instance_aggregator: typing.Callable = instance_aggregator
        if solver_aggregator is None:
            solver_aggregator = np.min if minimise else np.max
        self.solver_aggregator: typing.Callable = solver_aggregator
        self.minimise: bool = minimise
        self.post_process: typing.Callable = post_process
        self.use_time: UseTime = use_time
        self.metric = metric

    def __str__(self: SparkleObjective) -> str:
        """Return a stringified version."""
        return self.name

    @property
    def stem(self: SparkleObjective) -> str:
        """Return the stem of the objective name."""
        return self.name.split(":")[0]

    @property
    def time(self: SparkleObjective) -> bool:
        """Return whether the objective is time based."""
        return self.use_time != UseTime.NO


class PAR(SparkleObjective):
    """Penalised Averaged Runtime Objective for Sparkle."""

    negative_status = {
        SolverStatus.CRASHED,
        SolverStatus.KILLED,
        SolverStatus.ERROR,
        SolverStatus.TIMEOUT,
        SolverStatus.WRONG,
        SolverStatus.UNKNOWN,
    }

    def __init__(
        self: PAR, k: int = 10, minimise: bool = True, metric: bool = False
    ) -> None:
        """Initialize PAR."""
        self.k = k
        if k <= 0:
            raise ValueError("k must be greater than 0.")

        def penalise(value: float, cutoff: float, status: SolverStatus) -> float:
            """Return penalised value."""
            if status in PAR.negative_status or value > cutoff:
                return cutoff * self.k
            return value

        super().__init__(
            f"PAR{k}",
            minimise=minimise,
            use_time=UseTime.CPU_TIME,
            post_process=penalise,
            metric=metric,
        )



"""This package provides specific Configurator class implementations for Sparkle."""

from sparkle.configurator.configurator import Configurator
from sparkle.configurator.implementations.irace import IRACE, IRACEScenario
from sparkle.configurator.implementations.paramils import ParamILS, ParamILSScenario
from sparkle.configurator.implementations.smac2 import SMAC2, SMAC2Scenario
from sparkle.configurator.implementations.smac3 import SMAC3, SMAC3Scenario


def resolve_configurator(configurator_name: str) -> Configurator:
    """Returns the Configurator subclass by name."""
    subclass_names = [
        IRACE.__name__,
        ParamILS.__name__,
        SMAC2.__name__,
        SMAC3.__name__,
    ]
    if configurator_name in subclass_names:
        return eval(configurator_name)
    return None



#!/usr/bin/env python3
"""Sparkle output structures."""

from __future__ import annotations
from pathlib import Path

from runrunner.base import Status

from sparkle.solver import Solver
from sparkle.types import SolverStatus, SparkleObjective
from sparkle.instance import InstanceSet
from sparkle.structures import PerformanceDataFrame


class ValidationResults:
    """Class that stores validation information and results."""

    def __init__(
        self: ValidationResults,
        solver: Solver,
        configuration: dict,
        instance_set: InstanceSet,
        results: list[list[str, Status, float, float]],
    ) -> None:
        """Initalize ValidationResults.

        Args:
            solver: The name of the solver
            configuration: The configuration being used
            instance_set: The set of instances
            results: Validation results in the format:
                [["instance", "status", "quality", "runtime"]]
        """
        self.solver = solver
        self.configuration = configuration
        self.instance_set = instance_set
        self.result_header = ["instance", "status", "quality", "runtime"]
        self.result_vals = results


class SelectionSolverData:
    """Class that stores solver information."""

    def __init__(
        self: SelectionSolverData,
        solver_performance_ranking: list[tuple[str, float]],
        num_solvers: int,
    ) -> None:
        """Initalize SelectionSolverData.

        Args:
            solver_performance_ranking: list with solvers ranked by avg. performance
            num_solvers: The number of solvers
        """
        self.solver_performance_ranking = solver_performance_ranking
        self.single_best_solver = solver_performance_ranking[0][0]
        self.num_solvers = num_solvers


class SelectionPerformance:
    """Class that stores selection performance results."""

    def __init__(
        self: SelectionSolverData,
        performance_path: Path,
        vbs_performance: float,
        objective: SparkleObjective,
    ) -> None:
        """Initalize SelectionPerformance.

        Args:
            performance_path: Path to portfolio selector performance
            vbs_performance: The performance of the virtual best selector
            objective: The objective (Performance type)
        """
        performance_data = PerformanceDataFrame(performance_path)
        from sparkle.selector import SelectionScenario

        self.actual_performance_data = performance_data.get_value(
            solver=SelectionScenario.__selector_solver_name__, objective=objective.name
        )
        self.vbs_performance = vbs_performance
        self.actual_performance = performance_data.mean(objective=objective.name)
        self.metric = objective.name


class ParallelPortfolioResults:
    """Class that stores parallel portfolio results."""

    def __init__(
        self: ParallelPortfolioResults,
        unsolved_instances: int,
        sbs: str,
        runtime_solvers: dict[str, float],
        instance_results: dict[str, list],
    ) -> None:
        """Initalize SelectionSolverData.

        Args:
            unsolved_instances: Number of unsolved instances
            sbs: Name of the single best solver
            runtime_solvers: Dictionary containing penalised average runtime per solver
            instance_results: Dictionary containing
        """
        self.unsolved_instances = unsolved_instances
        self.sbs = sbs
        self.runtime_solvers = runtime_solvers
        self.instance_results = instance_results

        self.solver_performance = {}
        # Iterate over each instance and aggregate the results
        for _, results in self.instance_results.items():
            for solver_result in results:
                solver_name = solver_result[0]
                outcome = solver_result[1]
                # Initialize the solver's record in solver_performance if not present
                if solver_name not in self.solver_performance:
                    self.solver_performance[solver_name] = {
                        status: 0 for status in SolverStatus
                    }
                # Increment the appropriate outcome count
                self.solver_performance[solver_name][outcome] += 1



"""Sparkle class to organise configuration output."""

from __future__ import annotations

from sparkle.structures import PerformanceDataFrame
from sparkle.instance import InstanceSet
from sparkle.configurator.configurator import ConfigurationScenario
from sparkle.types import SparkleObjective, SolverStatus


class ConfigurationResult:
    """Class that represents result of configuration on an instance set."""

    def __init__(
        self: ConfigurationResult,
        instance_set: str,
        default_instance_performance: list[float],
        best_instance_performance: list[float],
        instance_status_default: dict[str, SolverStatus],
        instance_status_best: dict[str, SolverStatus],
        objective: SparkleObjective,
    ) -> None:
        """Initialize a configuration result.

        All input sequences are the results per instance.

        Args:
            instance_set: The name of the instance set
            default_instance_performance: The default instance performance
            best_instance_performance: The best instance performance
            performance: The performance of the configuration
            instance_status_default: The status of the default configuration
            instance_status_best: The status of the best configuration
            objective: The objective
        """
        self.default_instance_performance = default_instance_performance
        self.default_performance: float = objective.instance_aggregator(
            default_instance_performance
        )
        self.best_instance_performance = best_instance_performance
        self.best_performance: float = objective.instance_aggregator(
            best_instance_performance
        )
        self.instance_status_default = instance_status_default
        self.instance_status_best = instance_status_best
        self.instance_set_name = instance_set

    def serialise(self: ConfigurationResult) -> dict[str, float | list[float]]:
        """Serialise the data."""
        return {
            "instance_set": self.instance_set_name,
            "default_performance": self.default_performance,
            "best_performance": self.best_performance,
            "default_instance_performance": self.default_instance_performance,
            "best_instance_performance": self.best_instance_performance,
            "instance_status_default": self.instance_status_default,
            "instance_status_best": self.instance_status_best,
        }


class ConfigurationOutput:
    """Class that collects configuration data and outputs it a JSON format."""

    def __init__(
        self: ConfigurationOutput,
        config_scenario: ConfigurationScenario,
        performance_data: PerformanceDataFrame,
        possible_test_sets: list[InstanceSet] = None,
    ) -> None:
        """Initialize Configurator Output class.

        Args:
            config_scenario: The scenario to output
            performance_data: Performance data
            possible_test_sets: Instance Sets possibly used for testing
        """
        self.solver = config_scenario.solver
        self.configurator = config_scenario.configurator
        self.instance_set_train = config_scenario.instance_set

        # Filter data on this scenario
        performance_data_config = performance_data.clone()
        performance_data_config.remove_solver(
            [
                s
                for s in performance_data_config.solvers
                if s != str(self.solver.directory)
            ]
        )
        used_configs = config_scenario.configuration_ids + [
            PerformanceDataFrame.default_configuration
        ]
        removable = [
            c for c in performance_data_config.configuration_ids if c not in used_configs
        ]
        performance_data_config.remove_configuration(
            str(self.solver.directory), removable
        )
        self.test_instance_sets = []
        for test_set in possible_test_sets:
            if test_set.name == self.instance_set_train.name:
                continue
            for instance in test_set.instance_names:
                if (
                    instance not in performance_data_config.instances
                    or performance_data_config.is_missing(
                        str(self.solver.directory), instance
                    )
                ):
                    continue
            self.test_instance_sets.append(test_set)
        self.directory = config_scenario.directory
        self.config_scenario = config_scenario

        # Retrieve all configurations
        solver_key = str(self.solver.directory)
        config_keys = performance_data_config.get_configurations(solver_key)
        self.all_configurations = performance_data_config.get_full_configuration(
            solver_key, config_keys
        )

        # Retrieve configuration performances
        train_instances = self.instance_set_train.instance_names
        # Retrieve Default (No configuration) performance
        _, self.default_performance_train = (
            performance_data_config.configuration_performance(
                solver_key,
                PerformanceDataFrame.default_configuration,
                objective=self.config_scenario.sparkle_objectives[0],
                instances=train_instances,
            )
        )

        _, self.default_performance_per_instance_train = (
            performance_data_config.configuration_performance(
                solver_key,
                PerformanceDataFrame.default_configuration,
                objective=self.config_scenario.sparkle_objectives[0],
                instances=train_instances,
                per_instance=True,
            )
        )

        # Retrieve best found configuration
        self.best_configuration_key, self.best_performance_train = (
            performance_data_config.best_configuration(
                solver_key,
                objective=self.config_scenario.sparkle_objective,
                instances=train_instances,
            )
        )
        self.best_configuration = self.all_configurations[
            config_keys.index(self.best_configuration_key)
        ]

        # TODO keep all instance set performance data together in a dictionary instead
        # of variables for train and test
        # Shitty hack to get status objective
        status_objective = [
            o
            for o in performance_data_config.objective_names
            if o.lower().startswith("status")
        ][0]
        self.instance_set_results: dict[str, ConfigurationResult] = {}
        for instance_set in self.test_instance_sets + [self.instance_set_train]:
            instances = instance_set.instance_names
            _, default_performance_per_instance = (
                performance_data_config.configuration_performance(
                    solver_key,
                    PerformanceDataFrame.default_configuration,
                    objective=self.config_scenario.sparkle_objective,
                    instances=instances,
                    per_instance=True,
                )
            )
            _, best_conf_performance_per_instance = (
                performance_data_config.configuration_performance(
                    solver_key,
                    self.best_configuration_key,
                    objective=self.config_scenario.sparkle_objective,
                    instances=instances,
                    per_instance=True,
                )
            )
            instance_status_default = {
                str(i): performance_data_config.get_value(
                    solver_key,
                    configuration=PerformanceDataFrame.default_configuration,
                    objective=status_objective,
                    instance=[i],
                )
                for i in instances
            }
            instance_status_best_conf = {
                str(i): performance_data_config.get_value(
                    solver_key,
                    configuration=self.best_configuration_key,
                    objective=status_objective,
                    instance=[i],
                )
                for i in instances
            }
            self.instance_set_results[instance_set.name] = ConfigurationResult(
                instance_set.name,
                default_performance_per_instance,
                best_conf_performance_per_instance,
                instance_status_default,
                instance_status_best_conf,
                self.config_scenario.sparkle_objectives[0],
            )

    def serialise(self: ConfigurationOutput) -> dict:
        """Serialise the configuration output."""
        return {
            "solver": self.solver.name,
            "configurator": self.configurator.__name__,
            "best_configuration": self.best_configuration,
            "best_performance_train": self.best_performance_train,
            "scenario": {
                str(key): str(value)
                for key, value in self.config_scenario.serialise().items()
            },
        }



#!/usr/bin/env python3
"""Sparkle command to configure a solver."""

from __future__ import annotations
from pathlib import Path
import argparse
import sys

from runrunner import Runner

from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_object_name, resolve_instance_name
from sparkle.CLI.help import argparse_custom as ac

from sparkle.platform.settings_objects import Settings
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.solver import Solver
from sparkle.instance import Instance_Set


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Configure a solver in the platform.",
        epilog=(
            "Note that the test instance set is only used if the ``--ablation``"
            " or ``--validation`` flags are given"
        ),
    )
    parser.add_argument(*ac.SolverArgument.names, **ac.SolverArgument.kwargs)
    parser.add_argument(
        *ac.InstanceSetTrainArgument.names, **ac.InstanceSetTrainArgument.kwargs
    )
    parser.add_argument(
        *ac.InstanceSetTestArgument.names, **ac.InstanceSetTestArgument.kwargs
    )
    parser.add_argument(
        *ac.TestSetRunAllConfigurationArgument.names,
        **ac.TestSetRunAllConfigurationArgument.kwargs,
    )
    parser.add_argument(*ac.UseFeaturesArgument.names, **ac.UseFeaturesArgument.kwargs)
    # Settings Arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(
        *Settings.OPTION_configurator.args, **Settings.OPTION_configurator.kwargs
    )
    parser.add_argument(
        *Settings.OPTION_objectives.args, **Settings.OPTION_objectives.kwargs
    )
    parser.add_argument(
        *Settings.OPTION_solver_cutoff_time.args,
        **Settings.OPTION_solver_cutoff_time.kwargs,
    )
    parser.add_argument(
        *Settings.OPTION_configurator_solver_call_budget.args,
        **Settings.OPTION_configurator_solver_call_budget.kwargs,
    )
    parser.add_argument(
        *Settings.OPTION_configurator_number_of_runs.args,
        **Settings.OPTION_configurator_number_of_runs.kwargs,
    )
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the configure solver command."""
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(argsv=args)

    # Log command call
    sl.log_command(sys.argv, settings.random_state)
    check_for_initialise()

    configurator = settings.configurator
    # Check configurator is available
    if not configurator.check_requirements(verbose=True):
        print(
            f"{configurator.name} is not available. "
            "Please inspect possible warnings above."
        )
        print(f"Would you like to install {configurator.name}? (Y/n)")
        if input().lower().strip() == "y":
            configurator.download_requirements()
        else:
            sys.exit()
        if not configurator.check_requirements(verbose=True):
            raise RuntimeError(f"Failed to install {configurator.name}.")
        sys.exit(-1)

    # Compare current settings to latest.ini
    prev_settings = Settings(Path(Settings.DEFAULT_previous_settings_path))
    Settings.check_settings_changes(settings, prev_settings)

    configurator = gv.settings().configurator

    # Check configurator is available
    if not configurator.check_requirements(verbose=True):
        print(
            f"{configurator.name} is not available. "
            "Please inspect possible warnings above."
        )
        print(f"Would you like to install {configurator.name}? (Y/n)")
        if input().lower().strip() == "y":
            configurator.download_requirements()
        else:
            sys.exit()
        if not configurator.check_requirements(verbose=True):
            raise RuntimeError(f"Failed to install {configurator.name}.")
        sys.exit(-1)

    # Compare current settings to latest.ini
    prev_settings = Settings(Path(Settings.DEFAULT_previous_settings_path))
    Settings.check_settings_changes(gv.settings(), prev_settings)

    solver: Solver = resolve_object_name(
        args.solver,
        gv.file_storage_data_mapping[gv.solver_nickname_list_path],
        settings.DEFAULT_solver_dir,
        class_name=Solver,
    )
    if solver is None:
        raise ValueError(f"Solver {args.solver} not found.")
    instance_set_train = resolve_object_name(
        args.instance_set_train,
        gv.file_storage_data_mapping[gv.instances_nickname_path],
        settings.DEFAULT_instance_dir,
        Instance_Set,
    )
    if instance_set_train is None:
        raise ValueError(f"Instance set {args.instance_set_train} not found.")
    instance_set_test = args.instance_set_test
    if instance_set_test is not None:
        instance_set_test = resolve_object_name(
            args.instance_set_test,
            gv.file_storage_data_mapping[gv.instances_nickname_path],
            settings.DEFAULT_instance_dir,
            Instance_Set,
        )
    use_features = args.use_features
    run_on = settings.run_on

    configurator_settings = settings.get_configurator_settings(configurator.name)

    sparkle_objectives = settings.objectives
    if len(sparkle_objectives) > 1:
        print(
            f"WARNING: {configurator.name} does not have multi objective support. "
            f"Only the first objective ({sparkle_objectives[0]}) will be optimised."
        )

    performance_data = PerformanceDataFrame(settings.DEFAULT_performance_data_path)

    # Check if given objectives are in the data frame
    for objective in sparkle_objectives:
        if objective.name not in performance_data.objective_names:
            print(
                f"WARNING: Objective {objective.name} not found in performance data. "
                "Adding to data frame."
            )
            performance_data.add_objective(objective.name)

    if use_features:
        feature_data = FeatureDataFrame(settings.DEFAULT_feature_data_path)
        # Check that the train instance set is in the feature data frame
        invalid = False
        remaining_instance_jobs = set(
            [instance for instance, _, _ in feature_data.remaining_jobs()]
        )
        for instance in instance_set_train.instance_paths:
            if str(instance) not in feature_data.instances:
                print(f"ERROR: Train Instance {instance} not found in feature data.")
                invalid = True
            elif instance in remaining_instance_jobs:  # Check jobs
                print(f"ERROR: Features have not been computed for instance {instance}.")
                invalid = True
        if invalid:
            sys.exit(-1)
        configurator_settings.update({"feature_data": feature_data})

    number_of_runs = settings.configurator_number_of_runs
    output_path = settings.get_configurator_output_path(configurator)
    config_scenario = configurator.scenario_class()(
        solver,
        instance_set_train,
        sparkle_objectives,
        number_of_runs,
        output_path,
        **configurator_settings,
    )

    # Run the default configuration
    default_jobs = [
        (solver, config_id, instance, run_id)
        for solver, config_id, instance, run_id in performance_data.get_job_list()
        if config_id == PerformanceDataFrame.default_configuration
    ]

    sbatch_options = settings.sbatch_settings
    slurm_prepend = settings.slurm_job_prepend
    dependency_job_list = configurator.configure(
        scenario=config_scenario,
        data_target=performance_data,
        sbatch_options=sbatch_options,
        slurm_prepend=slurm_prepend,
        num_parallel_jobs=settings.slurm_jobs_in_parallel,
        base_dir=sl.caller_log_dir,
        run_on=run_on,
    )

    # If we have default configurations that need to be run, schedule them too
    if default_jobs:
        # Edit jobs to incorporate file paths
        instances = []
        for _, _, instance, _ in default_jobs:
            instance_path = resolve_instance_name(
                instance, settings.DEFAULT_instance_dir
            )
            instances.append(instance_path)
        default_job = solver.run_performance_dataframe(
            instances,
            performance_data,
            PerformanceDataFrame.default_configuration,
            sbatch_options=sbatch_options,
            slurm_prepend=slurm_prepend,
            cutoff_time=config_scenario.solver_cutoff_time,
            log_dir=config_scenario.validation,
            base_dir=sl.caller_log_dir,
            job_name=f"Default Configuration: {solver.name} Validation on "
            f"{instance_set_train.name}",
            run_on=run_on,
        )
        dependency_job_list.append(default_job)

    if instance_set_test is not None:
        # Schedule test set jobs
        if args.test_set_run_all_configurations:
            # TODO: Schedule test set runs for all configurations
            print("Running all configurations on test set is not implemented yet.")
            pass
        else:
            # We place the results in the index we just added
            run_index = list(
                set(
                    [
                        performance_data.get_instance_num_runs(str(i))
                        for i in instance_set_test.instance_names
                    ]
                )
            )
            test_set_job = solver.run_performance_dataframe(
                instance_set_test,
                performance_data,
                run_ids=run_index,
                cutoff_time=config_scenario.solver_cutoff_time,
                objective=config_scenario.sparkle_objective,
                train_set=instance_set_train,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                log_dir=config_scenario.validation,
                base_dir=sl.caller_log_dir,
                dependencies=dependency_job_list,
                job_name=f"Best Configuration: {solver.name} Validation on "
                f"{instance_set_test.name}",
                run_on=run_on,
            )
            dependency_job_list.append(test_set_job)

    if run_on == Runner.SLURM:
        job_id_str = ",".join([run.run_id for run in dependency_job_list])
        print(
            f"Running {configurator.name} configuration through Slurm with job "
            f"id(s): {job_id_str}"
        )
    else:
        print("Running configuration finished!")

    # Write used settings to file
    settings.write_used_settings()
    # Write used scenario to file
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
"""Sparkle command to compute features for instances."""

from __future__ import annotations
import sys
import argparse
from pathlib import Path

from runrunner.base import Run, Runner

from sparkle.selector import Extractor
from sparkle.platform.settings_objects import Settings
from sparkle.structures import FeatureDataFrame
from sparkle.instance import Instance_Set, InstanceSet


from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help.nicknames import resolve_instance_name


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Sparkle command to Compute features "
        "for instances using added extractors "
        "and instances."
    )
    parser.add_argument(
        *ac.RecomputeFeaturesArgument.names, **ac.RecomputeFeaturesArgument.kwargs
    )
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def compute_features(
    feature_data: Path | FeatureDataFrame,
    recompute: bool,
    run_on: Runner = Runner.SLURM,
) -> list[Run]:
    """Compute features for all instance and feature extractor combinations.

    A RunRunner run is submitted for the computation of the features.
    The results are then stored in the csv file specified by feature_data_csv_path.

    Args:
        feature_data: Feature Data Frame to use, or path to read it from.
        recompute: Specifies if features should be recomputed.
        run_on: Runner
            On which computer or cluster environment to run the solvers.
            Available: Runner.LOCAL, Runner.SLURM. Default: Runner.SLURM

    Returns:
        The Slurm job or Local job
    """
    if isinstance(feature_data, Path):
        feature_data = FeatureDataFrame(feature_data)
    if recompute:
        feature_data.reset_dataframe()
    jobs = feature_data.remaining_jobs()

    # Lookup all instances to resolve the instance paths later
    instances: list[InstanceSet] = []
    for instance_dir in gv.settings().DEFAULT_instance_dir.iterdir():
        if instance_dir.is_dir():
            instances.append(Instance_Set(instance_dir))

    # If there are no jobs, stop
    if not jobs:
        print(
            "No feature computation jobs to run; stopping execution! To recompute "
            "feature values use the --recompute flag."
        )
        return None
    cutoff = gv.settings().extractor_cutoff_time
    instance_paths = set()
    grouped_job_list: dict[str, dict[str, list[str]]] = {}

    # Group the jobs by extractor/feature group
    for instance_name, extractor_name, feature_group in jobs:
        if extractor_name not in grouped_job_list:
            grouped_job_list[extractor_name] = {}
        if feature_group not in grouped_job_list[extractor_name]:
            grouped_job_list[extractor_name][feature_group] = []
        instance_path = resolve_instance_name(str(instance_name), instances)
        grouped_job_list[extractor_name][feature_group].append(instance_path)

    sbatch_options = gv.settings().sbatch_settings
    slurm_prepend = gv.settings().slurm_job_prepend
    srun_options = ["-N1", "-n1"] + sbatch_options
    runs = []
    for extractor_name, feature_groups in grouped_job_list.items():
        extractor_path = gv.settings().DEFAULT_extractor_dir / extractor_name
        extractor = Extractor(extractor_path)
        for feature_group, instance_paths in feature_groups.items():
            run = extractor.run_cli(
                instance_paths,
                feature_data,
                cutoff,
                feature_group if extractor.groupwise_computation else None,
                run_on,
                sbatch_options,
                srun_options,
                gv.settings().slurm_jobs_in_parallel,
                slurm_prepend,
                log_dir=sl.caller_log_dir,
            )
            runs.append(run)
    return runs


def main(argv: list[str]) -> None:
    """Main function of the compute features command."""
    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)
    run_on = settings.run_on

    # Log command call
    sl.log_command(sys.argv, settings.random_state)
    check_for_initialise()

    # Check if there are any feature extractors registered
    if not any([p.is_dir() for p in gv.settings().DEFAULT_extractor_dir.iterdir()]):
        print(
            "No feature extractors present! Add feature extractors to Sparkle "
            "by using the add_feature_extractor command."
        )
        sys.exit()

    # Start compute features
    print("Start computing features ...")
    compute_features(settings.DEFAULT_feature_data_path, args.recompute, run_on)

    # Write used settings to file
    gv.settings().write_used_settings()
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Helper functions for CLI nicknames."""

from __future__ import annotations
from pathlib import Path
from typing import Callable
from sparkle.instance import Instance_Set, InstanceSet


def resolve_object_name(
    name: str | Path,
    nickname_dict: dict = {},
    target_dir: Path = Path(),
    class_name: Callable = None,
) -> Path | any:
    """Attempts to resolve a (nick) name.

    Args:
        name: The (nick)name to resolve
        target_dir: The location where the file object should exist
        nickname_dict: Nicknames
        class_name: If passed, will attempt to return an object
            that is constructed from this Path.

    Returns:
        Path to the object, None if unresolvable.
    """
    path = None
    # We cannot handle None as a name
    if name is None:
        return None
    # First check if the name already is a path
    if isinstance(name, (str, Path)) and Path(name).exists():
        path = Path(name)
    # Second check if its a nickname registered in Sparkle
    elif str(name) in nickname_dict:
        path = Path(nickname_dict[str(name)])
    # Third check if we can create a valid path with the name
    elif isinstance(name, (str, Path)) and (target_dir / name).exists():
        path = target_dir / name
    # Finally, attempt to construct the object from the Path
    try:
        if class_name is not None:
            if path is not None:
                return class_name(path)
            if name is not None:
                return class_name(name)
    except Exception:
        return None
    return path


def resolve_instance_name(
    name: str, target: Path | list[InstanceSet], return_path: bool = True
) -> str | InstanceSet:
    """Attempts to resolve an instance name.

    Args:
        name: The name to resolve
        target: Path to look for instance sets or the instance sets directly
        return_path: Whether to return the path of the instance or the instance set

    Returns:
        The path or the instance set of the given instance name
    """
    # Check if name is a multi file instance path

    # Check if the name is already an instance file path
    name_path = Path(name)
    if name_path.exists() and name_path.is_file():
        return name
    # Attempt to find files
    matches = [p for p in name_path.parent.glob(name_path.name + ".*")]
    if matches:
        return " ".join(str(p) for p in matches)  # Concat for multi file instance
    # Target is a path to a directory that contains instance directories
    elif isinstance(target, Path):
        instances = []
        for instance_dir in target.iterdir():
            if instance_dir.is_dir():
                instances.append(Instance_Set(instance_dir))
        target = instances

    out_set = None
    for instance_set in target:
        instance_path = instance_set.get_path_by_name(name)
        if instance_path is None:
            continue
        out_set = instance_set
        # Handle multi file instance
        instance_path = (
            [instance_path] if not isinstance(instance_path, list) else instance_path
        )
        instance_path = " ".join(str(p) for p in instance_path)
        break

    return instance_path if return_path else out_set



#!/usr/bin/env python3
"""Sparkle command to add a solver to the Sparkle platform."""

import os
import stat
import sys
import argparse
import shutil
from pathlib import Path

from sparkle.tools.parameters import PCSConvention

from sparkle.platform import file_help as sfh
from sparkle.CLI.help import global_variables as gv
from sparkle.structures import PerformanceDataFrame
from sparkle.solver import Solver, verifiers
from sparkle.CLI.help import logging as sl
from sparkle.CLI.initialise import check_for_initialise
from sparkle.CLI.help import argparse_custom as ac


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(description="Add a solver to the Sparkle platform.")
    parser.add_argument(
        *ac.DeterministicArgument.names, **ac.DeterministicArgument.kwargs
    )
    parser.add_argument(
        *ac.SolutionVerifierArgument.names, **ac.SolutionVerifierArgument.kwargs
    )
    parser.add_argument(
        *ac.NicknameSolverArgument.names, **ac.NicknameSolverArgument.kwargs
    )
    parser.add_argument(*ac.SolverPathArgument.names, **ac.SolverPathArgument.kwargs)
    parser.add_argument(*ac.SkipChecksArgument.names, **ac.SkipChecksArgument.kwargs)
    parser.add_argument(*ac.NoCopyArgument.names, **ac.NoCopyArgument.kwargs)
    return parser


def main(argv: list[str]) -> None:
    """Main function of the command."""
    # Log command call
    sl.log_command(sys.argv, gv.settings().random_state)
    check_for_initialise()

    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    solver_source = Path(args.solver_path)
    deterministic = args.deterministic
    solution_verifier = args.solution_verifier

    if not solver_source.exists():
        print(f'Solver path "{solver_source}" does not exist!')
        sys.exit(-1)

    # Make sure it is pointing to the verifiers module
    if solution_verifier:
        if Path(solution_verifier).is_file():  # File verifier
            solution_verifier = (
                verifiers.SolutionFileVerifier.__name__,
                solution_verifier,
            )
        elif solution_verifier not in verifiers.mapping:
            print(f"ERROR: Unknown solution verifier {solution_verifier}!")
            sys.exit(-1)

    nickname = args.nickname

    if args.run_checks:
        print("Running checks...")
        solver = Solver(Path(solver_source))
        if solver.pcs_file is None:
            print(
                "None or multiple .pcs files found. Solver "
                "is not valid for configuration."
            )
        else:
            print(f"PCS file detected: {solver.pcs_file.name}. ", end="")
            if solver.read_pcs_file():
                print("Can read the pcs file.")
            else:
                print("WARNING: Can not read the provided pcs file format.")

        wrapper_path = solver.directory / solver.wrapper
        if not wrapper_path.is_file():
            print(
                f"ERROR: Solver {solver_source.name} does not have a solver wrapper "
                f"(Missing file {solver.wrapper})."
            )
            sys.exit(-1)
        elif not os.access(wrapper_path, os.X_OK):
            print(
                f"ERROR: Solver {solver_source.name} wrapper file {solver.wrapper} "
                f" does not have execution rights set!"
            )
            sys.exit(-1)

    # Start add solver
    solver_directory = gv.settings().DEFAULT_solver_dir / solver_source.name
    if solver_directory.exists():
        print(
            f"ERROR: Solver {solver_source.name} already exists! Can not add new solver."
        )
        sys.exit(-1)
    if args.no_copy:
        print(f"Creating symbolic link from {solver_source} to {solver_directory}...")
        if not os.access(solver_source, os.W_OK):
            raise PermissionError(
                "Sparkle does not have the right to write to the destination folder."
            )
        solver_directory.symlink_to(solver_source.absolute())
    else:
        print(f"Copying {solver_source.name} to platform...")
        solver_directory.mkdir(parents=True)
        shutil.copytree(solver_source, solver_directory, dirs_exist_ok=True)

    # Save the deterministic bool in the solver
    with (solver_directory / Solver.meta_data).open("w+") as fout:
        fout.write(str({"deterministic": deterministic, "verifier": solution_verifier}))

    # Add RunSolver executable to the solver
    runsolver_path = gv.settings().DEFAULT_runsolver_exec
    if runsolver_path.name in [file.name for file in solver_directory.iterdir()]:
        print(
            "Warning! RunSolver executable detected in Solver "
            f"{solver_source.name}. This will be replaced with "
            f"Sparkle's version of RunSolver. ({runsolver_path})"
        )

    if runsolver_path.exists():
        runsolver_target = solver_directory / runsolver_path.name
        shutil.copyfile(runsolver_path, runsolver_target)
        runsolver_target.chmod(runsolver_target.stat().st_mode | stat.S_IEXEC)
    else:
        print("Warning! RunSolver does not exists. Falling back to PyRunSolver.")

    performance_data = PerformanceDataFrame(
        gv.settings().DEFAULT_performance_data_path,
        objectives=gv.settings().objectives,
    )
    performance_data.add_solver(str(solver_directory))
    performance_data.save_csv()

    print(f"Adding solver {solver_source.name} done!")

    if nickname is not None:
        sfh.add_remove_platform_item(
            solver_directory,
            gv.solver_nickname_list_path,
            gv.file_storage_data_mapping[gv.solver_nickname_list_path],
            key=nickname,
        )

    solver = Solver(solver_directory)  # Recreate solver from its new directory
    if solver.pcs_file is not None:
        # Generate missing PCS files
        # TODO: Only generate missing files
        print("Generating missing PCS files...")
        solver.port_pcs(PCSConvention.IRACE)  # Create PCS file for IRACE
        print("Generating IRACE done!")
        solver.port_pcs(PCSConvention.ParamILS)  # Create PCS file for ParamILS
        print("Generating ParamILS done!")

    # Write used settings to file
    gv.settings().write_used_settings()
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""This package provides configurator support for Sparkle."""

from sparkle.configurator.configurator import (
    Configurator,
    ConfigurationScenario,
    AblationScenario,
)

from sparkle.configurator.implementations import (
    SMAC2,
    SMAC2Scenario,
    SMAC3,
    SMAC3Scenario,
    ParamILS,
    ParamILSScenario,
    IRACE,
    IRACEScenario,
)



"""Extensions of the ConfigSpace lib."""

from __future__ import annotations
from typing_extensions import override, Iterable
import ast

import numpy as np
from typing import Any
from ConfigSpace import ConfigurationSpace
from ConfigSpace.hyperparameters import Hyperparameter
from ConfigSpace.types import Array, Mask, f64

from ConfigSpace.conditions import (
    Condition,
    AndConjunction,
    OrConjunction,
    EqualsCondition,
    GreaterThanCondition,
    InCondition,
    LessThanCondition,
    NotEqualsCondition,
)

from ConfigSpace.forbidden import (
    ForbiddenGreaterThanRelation,
    ForbiddenLessThanRelation,
    ForbiddenClause,
    ForbiddenConjunction,
    ForbiddenRelation,
    ForbiddenInClause,
    ForbiddenEqualsClause,
    ForbiddenAndConjunction,
)

_SENTINEL = object()


def expression_to_configspace(
    expression: str | ast.Module,
    configspace: ConfigurationSpace,
    target_parameter: Hyperparameter = None,
) -> ForbiddenClause | Condition:
    """Convert a logic expression to ConfigSpace expression.

    Args:
        expression: The expression to convert.
        configspace: The ConfigSpace to use.
        target_parameter: For conditions, will parse the expression as a condition
            underwhich the parameter will be active.
    """
    if isinstance(expression, str):
        try:
            expression = ast.parse(expression)
        except Exception as e:
            raise ValueError(f"Could not parse expression: '{expression}', {e}")
    if isinstance(expression, ast.Module):
        expression = expression.body[0]
    return recursive_conversion(
        expression, configspace, target_parameter=target_parameter
    )


def recursive_conversion(
    item: ast.mod,
    configspace: ConfigurationSpace,
    target_parameter: Hyperparameter = None,
) -> ForbiddenClause | Condition:
    """Recursively parse the AST tree to a ConfigSpace expression.

    Args:
        item: The item to parse.
        configspace: The ConfigSpace to use.
        target_parameter: For conditions, will parse the expression as a condition
            underwhich the parameter will be active.

    Returns:
        A ConfigSpace expression
    """
    if isinstance(item, list):
        if len(item) > 1:
            raise ValueError(f"Can not parse list of elements: {item}.")
        item = item[0]
    if isinstance(item, ast.Expr):
        return recursive_conversion(item.value, configspace, target_parameter)
    if isinstance(item, ast.Name):  # Convert to hyperparameter
        hp = configspace.get(item.id)
        return hp if hp is not None else item.id
    if isinstance(item, ast.Constant):
        return item.value
    if (
        isinstance(item, ast.Tuple)
        or isinstance(item, ast.Set)
        or isinstance(item, ast.List)
    ):
        values = []
        for v in item.elts:
            if isinstance(v, ast.Constant):
                values.append(v.value)
            elif isinstance(v, ast.Name):  # Check if its a parameter
                if v.id in list(configspace.values()):
                    raise ValueError(
                        f"Only constants allowed in tuples. Found: {item.elts}"
                    )
                values.append(v.id)  # String value was interpreted as parameter
        return values
    if isinstance(item, ast.BinOp):
        raise NotImplementedError("Binary operations not supported by ConfigSpace.")
    if isinstance(item, ast.BoolOp):
        values = [
            recursive_conversion(v, configspace, target_parameter) for v in item.values
        ]
        if isinstance(item.op, ast.Or):
            if target_parameter:
                return OrConjunction(*values)
            return ForbiddenOrConjunction(*values)
        elif isinstance(item.op, ast.And):
            if target_parameter:
                return AndConjunction(*values)
            return ForbiddenAndConjunction(*values)
        else:
            raise ValueError(f"Unknown boolean operator: {item.op}")
    if isinstance(item, ast.Compare):
        if len(item.ops) > 1:
            raise ValueError(f"Only single comparisons allowed. Found: {item.ops}")
        left = recursive_conversion(item.left, configspace, target_parameter)
        right = recursive_conversion(item.comparators, configspace, target_parameter)
        operator = item.ops[0]
        if isinstance(left, Hyperparameter):  # Convert to HP type
            if isinstance(right, Iterable) and not isinstance(right, str):
                right = [type(left.default_value)(v) for v in right]
                if len(right) == 1 and not isinstance(operator, ast.In):
                    right = right[0]
            elif isinstance(right, int):
                right = type(left.default_value)(right)

        if isinstance(operator, ast.Lt):
            if target_parameter:
                return LessThanCondition(target_parameter, left, right)
            return ForbiddenLessThanRelation(left=left, right=right)
        if isinstance(operator, ast.LtE):
            if target_parameter:
                raise ValueError("LessThanEquals not supported for conditions.")
            return ForbiddenLessThanEqualsRelation(left=left, right=right)
        if isinstance(operator, ast.Gt):
            if target_parameter:
                return GreaterThanCondition(target_parameter, left, right)
            return ForbiddenGreaterThanRelation(left=left, right=right)
        if isinstance(operator, ast.GtE):
            if target_parameter:
                raise ValueError("GreaterThanEquals not supported for conditions.")
            return ForbiddenGreaterThanEqualsRelation(left=left, right=right)
        if isinstance(operator, ast.Eq):
            if target_parameter:
                return EqualsCondition(target_parameter, left, right)
            return ForbiddenEqualsClause(hyperparameter=left, value=right)
        if isinstance(operator, ast.In):
            if target_parameter:
                return InCondition(target_parameter, left, right)
            return ForbiddenInClause(hyperparameter=left, values=right)
        if isinstance(operator, ast.NotEq):
            if target_parameter:
                return NotEqualsCondition(target_parameter, left, right)
            raise ValueError("NotEq operator not supported for ForbiddenClauses.")
        # The following classes do not (yet?) exist in configspace
        if isinstance(operator, ast.NotIn):
            raise ValueError("NotIn operator not supported for ForbiddenClauses.")
        if isinstance(operator, ast.Is):
            raise NotImplementedError("Is operator not supported.")
        if isinstance(operator, ast.IsNot):
            raise NotImplementedError("IsNot operator not supported.")
    raise ValueError(f"Unsupported type: {item}")


class ForbiddenLessThanEqualsRelation(ForbiddenLessThanRelation):
    """A ForbiddenLessThanEquals relation between two hyperparameters."""

    _RELATION_STR = "LESSEQUAL"

    def __repr__(self: ForbiddenLessThanEqualsRelation) -> str:
        """Return a string representation of the ForbiddenLessThanEqualsRelation."""
        return f"Forbidden: {self.left.name} <= {self.right.name}"

    @override
    def is_forbidden_value(
        self: ForbiddenLessThanEqualsRelation, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        # Relation is always evaluated against actual value and not vector rep
        left = values.get(self.left.name, _SENTINEL)
        if left is _SENTINEL:
            return False

        right = values.get(self.right.name, _SENTINEL)
        if right is _SENTINEL:
            return False

        return left <= right  # type: ignore

    @override
    def is_forbidden_vector(
        self: ForbiddenLessThanEqualsRelation, vector: Array[f64]
    ) -> bool:
        """Check if the vector is forbidden."""
        # Relation is always evaluated against actual value and not vector rep
        left: f64 = vector[self.vector_ids[0]]  # type: ignore
        right: f64 = vector[self.vector_ids[1]]  # type: ignore
        if np.isnan(left) or np.isnan(right):
            return False
        return self.left.to_value(left) <= self.right.to_value(right)  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenLessThanEqualsRelation, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        left = arr[self.vector_ids[0]]
        right = arr[self.vector_ids[1]]
        valid = ~(np.isnan(left) | np.isnan(right))
        out = np.zeros_like(valid)
        out[valid] = self.left.to_value(left[valid]) <= self.right.to_value(right[valid])
        return out


class ForbiddenGreaterThanEqualsRelation(ForbiddenGreaterThanRelation):
    """A ForbiddenGreaterThanEquals relation between two hyperparameters."""

    _RELATION_STR = "GREATEREQUAL"

    def __repr__(self: ForbiddenGreaterThanEqualsRelation) -> str:
        """Return a string representation of the ForbiddenGreaterThanEqualsRelation."""
        return f"Forbidden: {self.left.name} >= {self.right.name}"

    @override
    def is_forbidden_value(
        self: ForbiddenGreaterThanEqualsRelation, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        left = values.get(self.left.name, _SENTINEL)
        if left is _SENTINEL:
            return False

        right = values.get(self.right.name, _SENTINEL)
        if right is _SENTINEL:
            return False

        return left >= right  # type: ignore

    @override
    def is_forbidden_vector(
        self: ForbiddenGreaterThanEqualsRelation, vector: Array[f64]
    ) -> bool:
        """Check if the vector is forbidden."""
        # Relation is always evaluated against actual value and not vector rep
        left: f64 = vector[self.vector_ids[0]]  # type: ignore
        right: f64 = vector[self.vector_ids[1]]  # type: ignore
        if np.isnan(left) or np.isnan(right):
            return False
        return self.left.to_value(left) >= self.right.to_value(right)  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenGreaterThanEqualsRelation, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        left = arr[self.vector_ids[0]]
        right = arr[self.vector_ids[1]]
        valid = ~(np.isnan(left) | np.isnan(right))
        out = np.zeros_like(valid)
        out[valid] = self.left.to_value(left[valid]) >= self.right.to_value(right[valid])
        return out


class ForbiddenGreaterThanClause(ForbiddenEqualsClause):
    """A ForbiddenGreaterThanClause.

    It forbids a value from the value range of a hyperparameter to be
    *greater than* `value`.

    Forbids the value of the hyperparameter *a* to be greater than 2

    Args:
        hyperparameter: Methods on which a restriction will be made
        value: forbidden value
    """

    def __repr__(self: ForbiddenGreaterThanClause) -> str:
        """Return a string representation of the ForbiddenGreaterThanClause."""
        return f"Forbidden: {self.hyperparameter.name} > {self.value!r}"

    @override
    def is_forbidden_value(
        self: ForbiddenGreaterThanClause, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        return (  # type: ignore
            values.get(self.hyperparameter.name, _SENTINEL) > self.value
        )

    @override
    def is_forbidden_vector(
        self: ForbiddenGreaterThanClause, vector: Array[f64]
    ) -> bool:
        """Check if the vector is forbidden."""
        return vector[self.vector_id] > self.vector_value  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenGreaterThanClause, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        return np.greater(arr[self.vector_id], self.vector_value, dtype=np.bool_)

    @override
    def to_dict(self: ForbiddenGreaterThanClause) -> dict[str, Any]:
        """Convert the ForbiddenGreaterThanClause to a dictionary."""
        return {
            "name": self.hyperparameter.name,
            "type": "GREATER",
            "value": self.value,
        }


class ForbiddenGreaterEqualsClause(ForbiddenEqualsClause):
    """A ForbiddenGreaterEqualsClause.

    It forbids a value from the value range of a hyperparameter to be
    *greater or equal to* `value`.

    Forbids the value of the hyperparameter *a* to be greater or equal to 2

    Args:
        hyperparameter: Methods on which a restriction will be made
        value: forbidden value
    """

    def __repr__(self: ForbiddenGreaterEqualsClause) -> str:
        """Return a string representation of the ForbiddenGreaterEqualsClause."""
        return f"Forbidden: {self.hyperparameter.name} >= {self.value!r}"

    @override
    def is_forbidden_value(
        self: ForbiddenGreaterEqualsClause, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        return (  # type: ignore
            values.get(self.hyperparameter.name, _SENTINEL) >= self.value
        )

    @override
    def is_forbidden_vector(
        self: ForbiddenGreaterEqualsClause, vector: Array[f64]
    ) -> bool:
        """Check if the vector is forbidden."""
        return vector[self.vector_id] >= self.vector_value  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenGreaterEqualsClause, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        return np.greater_equal(arr[self.vector_id], self.vector_value, dtype=np.bool_)

    @override
    def to_dict(self: ForbiddenGreaterEqualsClause) -> dict[str, Any]:
        """Convert the ForbiddenGreaterEqualsClause to a dictionary."""
        return {
            "name": self.hyperparameter.name,
            "type": "GREATEREQUAL",
            "value": self.value,
        }


class ForbiddenLessThanClause(ForbiddenEqualsClause):
    """A ForbiddenLessThanClause.

    It forbids a value from the value range of a hyperparameter to be
    *less than* `value`.

    Args:
        hyperparameter: Methods on which a restriction will be made
        value: forbidden value
    """

    def __repr__(self: ForbiddenLessThanClause) -> str:
        """Return a string representation of the ForbiddenLessThanClause."""
        return f"Forbidden: {self.hyperparameter.name} < {self.value!r}"

    @override
    def is_forbidden_value(
        self: ForbiddenLessThanClause, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        return (  # type: ignore
            values.get(self.hyperparameter.name, _SENTINEL) < self.value
        )

    @override
    def is_forbidden_vector(self: ForbiddenLessThanClause, vector: Array[f64]) -> bool:
        """Check if the vector is forbidden."""
        return vector[self.vector_id] < self.vector_value  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenLessThanClause, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        return np.less(arr[self.vector_id], self.vector_value, dtype=np.bool_)

    @override
    def to_dict(self: ForbiddenLessThanClause) -> dict[str, Any]:
        """Convert the ForbiddenLessThanClause to a dictionary."""
        return {
            "name": self.hyperparameter.name,
            "type": "LESS",
            "value": self.value,
        }


class ForbiddenLessEqualsClause(ForbiddenEqualsClause):
    """A ForbiddenLessEqualsClause.

    It forbids a value from the value range of a hyperparameter to be
    *less or equal to* `value`.

    Args:
        hyperparameter: Methods on which a restriction will be made
        value: forbidden value
    """

    def __repr__(self: ForbiddenLessEqualsClause) -> str:
        """Return a string representation of the ForbiddenLessEqualsClause."""
        return f"Forbidden: {self.hyperparameter.name} <= {self.value!r}"

    @override
    def is_forbidden_value(
        self: ForbiddenLessEqualsClause, values: dict[str, Any]
    ) -> bool:
        """Check if the value is forbidden."""
        return (  # type: ignore
            values.get(self.hyperparameter.name, _SENTINEL) <= self.value
        )

    @override
    def is_forbidden_vector(self: ForbiddenLessEqualsClause, vector: Array[f64]) -> bool:
        """Check if the vector is forbidden."""
        return vector[self.vector_id] <= self.vector_value  # type: ignore

    @override
    def is_forbidden_vector_array(
        self: ForbiddenLessEqualsClause, arr: Array[f64]
    ) -> Mask:
        """Check if the vector array is forbidden."""
        return np.greater_equal(arr[self.vector_id], self.vector_value, dtype=np.bool_)

    @override
    def to_dict(self: ForbiddenLessEqualsClause) -> dict[str, Any]:
        """Convert the ForbiddenLessEqualsClause to a dictionary."""
        return {
            "name": self.hyperparameter.name,
            "type": "LESSEQUAL",
            "value": self.value,
        }


class ForbiddenOrConjunction(ForbiddenConjunction):
    """A ForbiddenOrConjunction.

    The ForbiddenOrConjunction combines forbidden-clauses, which allows to
    build powerful constraints.

    ```python exec="true", source="material-block" result="python"
    from ConfigSpace import (
        ConfigurationSpace,
        ForbiddenEqualsClause,
        ForbiddenInClause,
    )
    from sparkle.tools.configspace import ForbiddenOrConjunction

    cs = ConfigurationSpace({"a": [1, 2, 3], "b": [2, 5, 6]})
    forbidden_clause_a = ForbiddenEqualsClause(cs["a"], 2)
    forbidden_clause_b = ForbiddenInClause(cs["b"], [2])

    forbidden_clause = ForbiddenOrConjunction(forbidden_clause_a, forbidden_clause_b)

    cs.add(forbidden_clause)
    print(cs)
    ```

    Args:
        *args: forbidden clauses, which should be combined
    """

    components: tuple[ForbiddenClause | ForbiddenConjunction | ForbiddenRelation, ...]
    """Components of the conjunction."""

    dlcs: tuple[ForbiddenClause | ForbiddenRelation, ...]
    """Descendant literal clauses of the conjunction.

    These are the base forbidden clauses/relations that are part of conjunctions.

    !!! note

        This will only store a unique set of the descendant clauses, no duplicates.
    """

    def __repr__(self: ForbiddenOrConjunction) -> str:
        """Return a string representation of the ForbiddenOrConjunction."""
        return "(" + " || ".join([str(c) for c in self.components]) + ")"

    @override
    def is_forbidden_value(self: ForbiddenOrConjunction, values: dict[str, Any]) -> bool:
        """Check if the value is forbidden."""
        return any(
            [forbidden.is_forbidden_value(values) for forbidden in self.components]
        )

    @override
    def is_forbidden_vector(self: ForbiddenOrConjunction, vector: Array[f64]) -> bool:
        """Check if the vector is forbidden."""
        return any(
            forbidden.is_forbidden_vector(vector) for forbidden in self.components
        )

    @override
    def is_forbidden_vector_array(self: ForbiddenOrConjunction, arr: Array[f64]) -> Mask:
        """Check if the vector array is forbidden."""
        forbidden_mask: Mask = np.zeros(shape=arr.shape[1], dtype=np.bool_)
        for forbidden in self.components:
            forbidden_mask |= forbidden.is_forbidden_vector_array(arr)

        return forbidden_mask

    @override
    def to_dict(self: ForbiddenOrConjunction) -> dict[str, Any]:
        """Convert the ForbiddenOrConjunction to a dictionary."""
        return {
            "type": "OR",
            "clauses": [component.to_dict() for component in self.components],
        }



#!/usr/bin/env python3
"""Command to initialise a Sparkle platform."""

import subprocess
import argparse
import shutil
import os
import sys
import warnings
from pathlib import Path

from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help import snapshot_help as snh
from sparkle.CLI.help import global_variables as gv
from sparkle.configurator.implementations import IRACE, SMAC2, ParamILS
from sparkle.platform import Settings
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame


def parser_function() -> argparse.ArgumentParser:
    """Parse CLI arguments for the initialise command."""
    parser = argparse.ArgumentParser(
        description="Initialise the Sparkle platform in the current directory."
    )
    parser.add_argument(
        *ac.DownloadExamplesArgument.names, **ac.DownloadExamplesArgument.kwargs
    )
    parser.add_argument(
        *ac.NoSavePlatformArgument.names, **ac.NoSavePlatformArgument.kwargs
    )
    parser.add_argument(
        *ac.RebuildRunsolverArgument.names, **ac.RebuildRunsolverArgument.kwargs
    )
    return parser


def detect_sparkle_platform_exists(check: callable = all) -> Path:
    """Return whether a Sparkle platform is currently active.

    The default working directories are checked for existence, for each directory in the
    CWD. If any of the parents of the CWD yield true, this path is returned

    Args:
        check: Method to check if the working directory exists. Defaults to all.

    Returns:
      Path to the Sparkle platform if it exists, None otherwise.
    """
    cwd = Path.cwd()
    while str(cwd) != cwd.root:
        if check([(cwd / wd).exists() for wd in Settings.DEFAULT_working_dirs]):
            return cwd
        cwd = cwd.parent
    return None


def check_for_initialise() -> None:
    """Function to check if initialize command was executed and execute it otherwise.

    Args:
        argv: List of the arguments from the caller.
        requirements: The requirements that have to be executed before the calling
            function.
    """
    platform_path = detect_sparkle_platform_exists()
    if platform_path is None:
        print("-----------------------------------------------")
        print(
            "No Sparkle platform found; "
            "The platform will now be initialized automatically."
        )
        print("-----------------------------------------------")
        initialise_sparkle()
    elif platform_path != Path.cwd():
        print(
            f"[WARNING] Sparkle platform found in {platform_path} instead of "
            f"{Path.cwd()}. Switching to CWD to {platform_path}"
        )
        os.chdir(platform_path)


def initialise_sparkle(
    save_existing_platform: bool = True,
    interactive: bool = False,
    download_examples: bool = False,
    rebuild_runsolver: bool = False,
) -> None:
    """Initialize a new Sparkle platform.

    Args:
        save_existing_platform: If present, save the current platform as a snapshot.
        interactive: Ask for user input or not.
        download_examples: Downloads examples from the Sparkle Github.
            WARNING: May take a some time to complete due to the large amount of data.
        rebuild_runsolver: Will clean the RunSolver executable and rebuild it.
    """
    print("Start initialising Sparkle platform ...")
    if detect_sparkle_platform_exists(check=all):
        print("Current Sparkle platform found!")
        if save_existing_platform:
            print("Saving as snapshot...")
            snh.save_current_platform()
        snh.remove_current_platform(filter=[Settings.DEFAULT_settings_dir])
        print("Your settings directory was not removed.")

    for working_dir in Settings.DEFAULT_working_dirs:
        working_dir.mkdir(exist_ok=True)

    # Check if Settings file exists, otherwise initialise a default one
    if not Path(Settings.DEFAULT_settings_path).exists():
        print("Settings file does not exist, initializing default settings ...")
        gv.__settings = Settings(Settings.DEFAULT_example_settings_path)
        gv.settings().write_settings_ini(Path(Settings.DEFAULT_settings_path))

    # Initialise the FeatureDataFrame
    FeatureDataFrame(Settings.DEFAULT_feature_data_path)

    # Initialise the Performance DF with the static dimensions
    # TODO: We have many sparkle settings values regarding ``number of runs''
    # E.g. configurator, parallel portfolio, and here too. Should we unify this more, or
    # just make another setting that does this specifically for performance data?
    PerformanceDataFrame(
        Settings.DEFAULT_performance_data_path,
        objectives=gv.settings().objectives,
        n_runs=1,
    )

    if rebuild_runsolver:
        print("Cleaning Runsolver ...")
        runsolver_clean = subprocess.run(
            ["make", "clean"], cwd=Settings.DEFAULT_runsolver_dir, capture_output=True
        )
        if runsolver_clean.returncode != 0:
            warnings.warn(
                f"[{runsolver_clean.returncode}] Cleaning of Runsolver failed "
                f"with the following msg: {runsolver_clean.stdout.decode()}"
            )

    # Check that Runsolver is compiled, otherwise, compile
    if not Settings.DEFAULT_runsolver_exec.exists():
        print("Runsolver does not exist, trying to compile...")
        if not (Settings.DEFAULT_runsolver_dir / "Makefile").exists():
            warnings.warn(
                "Runsolver executable doesn't exist and cannot find makefile."
                " Please verify the contents of the directory: "
                f"{Settings.DEFAULT_runsolver_dir}"
            )
        else:
            compile_runsolver = subprocess.run(
                ["make"], cwd=Settings.DEFAULT_runsolver_dir, capture_output=True
            )
            if compile_runsolver.returncode != 0:
                warnings.warn(
                    "Compilation of Runsolver failed with the following msg:"
                    f"[{compile_runsolver.returncode}] "
                    f"{compile_runsolver.stderr.decode()}"
                )
            else:
                print("Runsolver compiled successfully!")

    # If Runsolver is compiled, check that it can be executed
    if Settings.DEFAULT_runsolver_exec.exists():
        runsolver_check = subprocess.run(
            [Settings.DEFAULT_runsolver_exec, "--version"], capture_output=True
        )
        if runsolver_check.returncode != 0:
            print(
                "WARNING: Runsolver executable cannot be run successfully. "
                "Please verify the following error messages:\n"
                f"{runsolver_check.stderr.decode()}"
            )

    # Check that java is available for SMAC2
    if shutil.which("java") is None:
        # NOTE: An automatic resolution of Java at this point would be good
        # However, loading modules from Python has thusfar not been successfull.
        print(
            "WARNING: Could not find Java as an executable! Java 1.8.0_402 is required"
            " to use SMAC2 or ParamILS as a configurator. Consider installing Java."
        )

    # Check for each configurator that it is available
    if interactive and not SMAC2.configurator_executable.exists():
        print("SMAC2 is not installed, would you like to install? (Y/n) ...")
        if input().lower() == "y":
            print("Installing SMAC2 ...")
            SMAC2.download_requirements()
    if interactive and not ParamILS.configurator_executable.exists():
        print("ParamILS is not installed, would you like to install? (Y/n) ...")
        if input().lower() == "y":
            print("Installing ParamILS ...")
            ParamILS.download_requirements()
    if interactive and not IRACE.check_requirements():
        if shutil.which("R") is None:
            print(
                "R is not installed, which is required for the IRACE "
                "configurator (installation). Consider installing R."
            )
        else:
            print("IRACE is not installed, would you like to install? (Y/n) ...")
            if input().lower() == "y":
                print("Installing IRACE ...")
                IRACE.download_requirements()
                irace = IRACE()
                print(f"Installed IRACE version {irace.version}")

    if download_examples:
        # Download Sparkle examples from Github
        print("Downloading examples ...")
        curl = subprocess.Popen(
            ["curl", "https://codeload.github.com/ADA-research/Sparkle/tar.gz/main"],
            stdout=subprocess.PIPE,
        )
        outpath = Path("outfile.tar.gz")
        with curl.stdout, outpath.open("wb") as outfile:
            tar = subprocess.Popen(
                ["tar", "-xz", "--strip=1", "Sparkle-main/Examples"],
                stdin=curl.stdout,
                stdout=outfile,
            )
        curl.wait()  # Wait for the download to complete
        tar.wait()  # Wait for the extraction to complete
        outpath.unlink(missing_ok=True)

    print("New Sparkle platform initialised!")


def main(argv: list[str]) -> None:
    """Main function of the command."""
    # Define command line arguments
    parser = parser_function()
    # Process command line arguments
    args = parser.parse_args(argv)
    initialise_sparkle(
        save_existing_platform=args.no_save,
        interactive=True,
        download_examples=args.download_examples,
        rebuild_runsolver=args.rebuild_runsolver,
    )
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Helper types for command line interface."""

from enum import Enum
from typing import Type


class VerbosityLevel(Enum):
    """Enum of possible verbosity states."""

    QUIET = 0
    STANDARD = 1


class TEXT(Enum):
    """Class for ANSI text formatting."""

    RESET = "\033[0m"
    BOLD = "\033[1m"
    ITALIC = "\033[3m"
    UNDERLINE = "\033[4m"

    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"

    # BG = Background
    BG_BLACK = "\033[40m"
    BG_RED = "\033[41m"
    BG_GREEN = "\033[42m"
    BG_YELLOW = "\033[43m"
    BG_BLUE = "\033[44m"
    BG_MAGENTA = "\033[45m"
    BG_CYAN = "\033[46m"
    BG_WHITE = "\033[47m"

    @classmethod
    def format_text(cls: Type["TEXT"], formats: list[str], text: str) -> str:
        """Styles the string based on the provided formats."""
        start_format = "".join(format_.value for format_ in formats)
        end_format = cls.RESET.value
        return f"{start_format}{text}{end_format}"



#!/usr/bin/env python3
"""Sparkle command to construct a portfolio selector."""

import sys
import argparse

from runrunner.base import Runner

from sparkle.selector import Selector, SelectionScenario
from sparkle.instance import Instance_Set

from sparkle.platform.settings_objects import Settings
from sparkle.structures import PerformanceDataFrame, FeatureDataFrame
from sparkle.types import resolve_objective
from sparkle.CLI.help import global_variables as gv
from sparkle.CLI.help import logging as sl
from sparkle.CLI.help import argparse_custom as ac
from sparkle.CLI.help.nicknames import resolve_object_name, resolve_instance_name
from sparkle.CLI.initialise import check_for_initialise


def parser_function() -> argparse.ArgumentParser:
    """Define the command line arguments."""
    parser = argparse.ArgumentParser(
        description="Command to construct a portfolio selector over all known features "
        "solver performances."
    )
    parser.add_argument(*ac.SolversArgument.names, **ac.SolversArgument.kwargs)
    parser.add_argument(
        *ac.RecomputePortfolioSelectorArgument.names,
        **ac.RecomputePortfolioSelectorArgument.kwargs,
    )
    parser.add_argument(*ac.ObjectiveArgument.names, **ac.ObjectiveArgument.kwargs)
    parser.add_argument(
        *ac.SelectorAblationArgument.names, **ac.SelectorAblationArgument.kwargs
    )
    parser.add_argument(
        *ac.InstanceSetTrainOptionalArgument.names,
        **ac.InstanceSetTrainOptionalArgument.kwargs,
    )
    # Solver Configurations arguments
    configuration_group = parser.add_mutually_exclusive_group(required=False)
    configuration_group.add_argument(
        *ac.AllSolverConfigurationArgument.names,
        **ac.AllSolverConfigurationArgument.kwargs,
    )
    configuration_group.add_argument(
        *ac.BestSolverConfigurationArgument.names,
        **ac.BestSolverConfigurationArgument.kwargs,
    )
    configuration_group.add_argument(
        *ac.DefaultSolverConfigurationArgument.names,
        **ac.DefaultSolverConfigurationArgument.kwargs,
    )
    # TODO: Allow user to specify configuration ids to use
    # Settings arguments
    parser.add_argument(*ac.SettingsFileArgument.names, **ac.SettingsFileArgument.kwargs)
    parser.add_argument(
        *Settings.OPTION_minimum_marginal_contribution.args,
        **Settings.OPTION_minimum_marginal_contribution.kwargs,
    )
    parser.add_argument(*Settings.OPTION_run_on.args, **Settings.OPTION_run_on.kwargs)
    return parser


def judge_exist_remaining_jobs(
    feature_data: FeatureDataFrame, performance_data: PerformanceDataFrame
) -> bool:
    """Return whether there are remaining feature or performance computation jobs."""
    missing_features = feature_data.has_missing_vectors()
    missing_performances = performance_data.has_missing_values
    if missing_features:
        print(
            "There remain unperformed feature computation jobs! Please run: "
            "'sparkle compute features'"
        )
    if missing_performances:
        print(
            "There remain unperformed performance computation jobs! Please run:\n"
            "'sparkle cleanup --performance-data'\n"
            "to check for missing values in the logs, otherwise run:\n"
            "'sparkle run solvers --performance-data'\n"
            "to compute missing values."
        )
    if missing_features or missing_performances:
        print(
            "Please first execute all unperformed jobs before constructing Sparkle "
            "portfolio selector."
        )
        sys.exit(-1)


def main(argv: list[str]) -> None:
    """Main method of construct portfolio selector."""
    # Define command line arguments
    parser = parser_function()

    # Process command line arguments
    args = parser.parse_args(argv)
    settings = gv.settings(args)

    # Log command call
    sl.log_command(sys.argv, settings.random_state)
    check_for_initialise()

    flag_recompute_portfolio = args.recompute_portfolio_selector
    solver_ablation = args.solver_ablation

    if args.objective is not None:
        objective = resolve_objective(args.objective)
    else:
        objective = settings.objectives[0]
        print(
            "WARNING: No objective specified, defaulting to first objective from "
            f"settings ({objective})."
        )
    run_on = settings.run_on

    print("Start constructing Sparkle portfolio selector ...")
    selector = Selector(settings.selection_class, settings.selection_model)

    instance_set = None
    if args.instance_set_train is not None:
        instance_set = resolve_object_name(
            args.instance_set_train,
            gv.file_storage_data_mapping[gv.instances_nickname_path],
            gv.settings().DEFAULT_instance_dir,
            Instance_Set,
        )

    solver_cutoff_time = gv.settings().solver_cutoff_time
    extractor_cutoff_time = gv.settings().extractor_cutoff_time

    performance_data = PerformanceDataFrame(gv.settings().DEFAULT_performance_data_path)
    feature_data = FeatureDataFrame(gv.settings().DEFAULT_feature_data_path)

    # Check that the feature data actually contains features (extractors)
    if feature_data.num_features == 0:
        print(
            "ERROR: Feature data is empty! Please add a feature extractor and run "
            "'sparkle compute features' first."
        )
        sys.exit(-1)

    # Filter objective
    performance_data.remove_objective(
        [obj for obj in performance_data.objective_names if obj != objective.name]
    )
    if instance_set is not None:
        removable_instances = [
            i for i in performance_data.instances if i not in instance_set.instance_names
        ]
        performance_data.remove_instances(removable_instances)
        feature_data.remove_instances(removable_instances)

    if args.solvers is not None:
        solvers = args.solvers
        removeable_solvers = [s for s in performance_data.solvers if s not in solvers]
        performance_data.remove_solver(removeable_solvers)
    else:
        solvers = sorted(
            [str(s) for s in gv.settings().DEFAULT_solver_dir.iterdir() if s.is_dir()]
        )

    # Check what configurations should be considered
    if args.best_configuration:
        configurations = {
            s: performance_data.best_configuration(s, objective=objective)
            for s in solvers
        }
    elif args.default_configuration:
        configurations = {s: PerformanceDataFrame.default_configuration for s in solvers}
    else:
        configurations = {s: performance_data.get_configurations(s) for s in solvers}
        if not args.all_configurations:  # Take the only configuration
            if any(len(c) > 1 for c in configurations.values()):
                print("ERROR: More than one configuration for the following solvers:")
                for solver, config in configurations.items():
                    if len(config) > 1:
                        print(f"\t{solver}: {config} configurations")
                raise ValueError(
                    "Please set the --all-configurations flag if you wish to use more "
                    "than one configuration per solver."
                )
    for solver in solvers:
        removeable_configs = [
            c
            for c in performance_data.get_configurations(solver)
            if c not in configurations[solver]
        ]
        performance_data.remove_configuration(solver, removeable_configs)

    judge_exist_remaining_jobs(feature_data, performance_data)
    if feature_data.has_missing_value():
        print(
            "WARNING: Missing values in the feature data, will be imputed as the mean "
            "value of all other non-missing values! Imputing all missing values..."
        )
        feature_data.impute_missing_values()

    # Filter the scenario from Solver (Configurations) that do not meet the minimum marginal contribution on the training set
    if gv.settings().minimum_marginal_contribution > 0.0:
        print(
            f"Filtering the scenario from Solver (Configurations) with contribution < {gv.settings().minimum_marginal_contribution} ..."
        )
        for (
            solver,
            config_id,
            marginal_contribution,
            _,
        ) in performance_data.marginal_contribution(objective=objective):
            if marginal_contribution < gv.settings().minimum_marginal_contribution:
                print(f"\tRemoving {solver}, {config_id} [{marginal_contribution}]")
                performance_data.remove_configuration(solver, config_id)

    selection_scenario = SelectionScenario(
        gv.settings().DEFAULT_selection_output,
        selector,
        objective,
        performance_data,
        feature_data,
        solver_cutoff=solver_cutoff_time,
        extractor_cutoff=extractor_cutoff_time,
        ablate=solver_ablation,
    )

    if selection_scenario.selector_file_path.exists():
        if not flag_recompute_portfolio:
            print(
                "Portfolio selector already exists. "
                "Set the recompute flag to remove and reconstruct."
            )
            sys.exit(-1)
        # Delete all selectors
        selection_scenario.selector_file_path.unlink(missing_ok=True)
        if selection_scenario.ablation_scenarios:
            for scenario in selection_scenario.ablation_scenarios:
                scenario.selector_file_path.unlink(missing_ok=True)

    sbatch_options = gv.settings().sbatch_settings
    slurm_prepend = gv.settings().slurm_job_prepend
    selector_run = selector.construct(
        selection_scenario,
        run_on=run_on,
        sbatch_options=sbatch_options,
        slurm_prepend=slurm_prepend,
        base_dir=sl.caller_log_dir,
    )
    jobs = [selector_run]
    if run_on == Runner.LOCAL:
        print("Sparkle portfolio selector constructed!")
    else:
        print("Sparkle portfolio selector constructor running...")

    # Validate the selector to run on the given instances
    instances = [
        resolve_instance_name(instance, Settings.DEFAULT_instance_dir)
        for instance in performance_data.instances
    ]
    selector_validation = selector.run_cli(
        selection_scenario.scenario_file,
        instances,
        feature_data.csv_filepath,
        run_on=run_on,
        sbatch_options=sbatch_options,
        slurm_prepend=slurm_prepend,
        dependencies=[selector_run],
        log_dir=sl.caller_log_dir,
    )
    jobs.append(selector_validation)

    if solver_ablation:
        for ablated_scenario in selection_scenario.ablation_scenarios:
            # Construct the ablated selector
            ablation_run = selector.construct(
                ablated_scenario,
                run_on=run_on,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                base_dir=sl.caller_log_dir,
            )
            # Validate the ablated selector
            ablation_validation = selector.run_cli(
                ablated_scenario.scenario_file,
                instances,
                feature_data.csv_filepath,
                run_on=run_on,
                sbatch_options=sbatch_options,
                slurm_prepend=slurm_prepend,
                job_name=f"Selector Ablation: {ablated_scenario.directory.name} on {len(instances)} instances",
                dependencies=[ablation_run],
                log_dir=sl.caller_log_dir,
            )
            jobs.extend([ablation_run, ablation_validation])

    if run_on == Runner.LOCAL:
        for job in jobs:
            job.wait()
        selector_validation.wait()
        print("Selector validation done!")
    else:
        print(
            f"Running selector construction through Slurm with job id(s): "
            f"{', '.join([d.run_id for d in jobs])}"
        )

    # Write used settings to file
    gv.settings().write_used_settings()
    sys.exit(0)


if __name__ == "__main__":
    main(sys.argv[1:])



"""Custom helper class and functions to process CLI arguments with argparse."""

from __future__ import annotations
from pathlib import Path
from typing import Any


class ArgumentContainer:
    """Helper class for more convenient argument packaging and access."""

    def __init__(
        self: ArgumentContainer, names: list[str], kwargs: dict[str, Any]
    ) -> None:
        """Create an ArgumentContainer.

        Args:
            names: List of names for the contained argument. For positional arguments
                this will typically contain two, the first one starting with '-' and the
                second one starting with '--'.
            kwargs: Keyword arguments needed by the method parser.add_argument which adds
                the contained argument to a parser.
        """
        self.names = names
        self.kwargs = kwargs


AllJobsArgument = ArgumentContainer(
    names=["--all"],
    kwargs={"action": "store_true", "help": "use all known job ID(s) for the command"},
)

AllSolverConfigurationArgument = ArgumentContainer(
    names=["--all-configurations"],
    kwargs={
        "action": "store_true",
        "help": "use all known configurations for the command",
    },
)

AllConfigurationArgument = ArgumentContainer(
    names=["--all-configurations"],
    kwargs={"action": "store_true", "help": "use all known Solver configurations"},
)

BestConfigurationArgument = ArgumentContainer(
    names=["--best-configuration"],
    kwargs={
        "required": False,
        "nargs": "?",
        "type": Path,
        "const": True,
        "default": False,
        "help": "Paths to instance(s) or instanceset(s) over "
        "which to determine the best configuration. If "
        "empty, all known instances are used.",
    },
)

BestSolverConfigurationArgument = ArgumentContainer(
    names=["--best-configuration"],
    kwargs={
        "action": "store_true",
        "help": "use the best configurations for the command",
    },
)

CancelJobsArgument = ArgumentContainer(
    names=["--cancel"],
    kwargs={"action": "store_true", "help": "cancel the job(s) with the given ID(s)"},
)

CheckTypeArgument = ArgumentContainer(
    names=["type"],
    kwargs={
        "choices": [
            "extractor",
            "feature-extractor",
            "solver",
            "instance-set",
            "ExtractorFeature-Extractor",
            "Instance-Set",
            "Solver",
            "FeatureExtractor",
            "InstanceSet",
        ],
        "help": "type of the object to check",
    },
)

CheckPathArgument = ArgumentContainer(
    names=["path"], kwargs={"type": Path, "help": "path to the object to check"}
)

CleanupArgumentAll = ArgumentContainer(
    names=["--all"], kwargs={"action": "store_true", "help": "clean all output files"}
)

CleanupArgumentRemove = ArgumentContainer(
    names=["--remove"],
    kwargs={
        "action": "store_true",
        "help": "remove all files in the platform, including "
        "user data such as InstanceSets and Solvers",
    },
)

CleanupArgumentLogs = ArgumentContainer(
    names=["--logs", "--log"],
    kwargs={"action": "store_true", "help": "Clean log files from the platform"},
)

CleanUpPerformanceDataArgument = ArgumentContainer(
    names=["--performance-data"],
    kwargs={"action": "store_true", "help": "clean performance data from empty lines"},
)


ConfigurationArgument = ArgumentContainer(
    names=["--configuration"],
    kwargs={
        "required": False,
        "type": str,
        "nargs": "+",
        "help": "The indices of which configurations to use",
    },
)

DefaultSolverConfigurationArgument = ArgumentContainer(
    names=["--default-configuration"],
    kwargs={
        "action": "store_true",
        "help": "use the default configurations for the command",
    },
)

DeterministicArgument = ArgumentContainer(
    names=["--deterministic"],
    kwargs={
        "action": "store_true",
        "help": "Flag indicating the solver is deterministic",
    },
)

DownloadExamplesArgument = ArgumentContainer(
    names=["--download-examples"],
    kwargs={
        "action": "store_true",
        "default": False,
        "required": False,
        "help": "Download the Examples into the directory.",
    },
)

ExtractorPathArgument = ArgumentContainer(
    names=["extractor_path"],
    kwargs={
        "metavar": "extractor-path",
        "type": str,
        "help": "path or nickname of the feature extractor",
    },
)

GenerateJSONArgument = ArgumentContainer(
    names=["--only-json"],
    kwargs={
        "required": False,
        "default": False,
        "type": bool,
        "help": "if set to True, only generate machine readable output",
    },
)

InstancePathOptional = ArgumentContainer(
    names=["instance_path"],
    kwargs={
        "type": Path,
        "nargs": "?",
        "help": "Path to an instance to use for the command",
    },
)

InstanceSetPathsArgument = ArgumentContainer(
    names=[
        "--instance-path",
        "--instance-set-path",
        "--instance",
        "--instance-set",
        "--instances",
        "--instance-sets",
        "--instance-paths",
        "--instance-set-paths",
    ],
    kwargs={
        "required": False,
        "nargs": "+",
        "type": Path,
        "help": "Path to an instance (set)",
    },
)

InstanceSetRequiredArgument = ArgumentContainer(
    names=["--instance", "--instance-set"],
    kwargs={"required": True, "type": Path, "help": "path to instance (set)"},
)

InstanceSetTestArgument = ArgumentContainer(
    names=["--instance-set-test"],
    kwargs={"required": False, "type": Path, "help": "path to test instance set"},
)

InstanceSetTrainArgument = ArgumentContainer(
    names=["--instance-set-train"],
    kwargs={"required": True, "type": Path, "help": "path to training instance set"},
)

InstanceSetTestAblationArgument = ArgumentContainer(
    names=["--instance-set-test"],
    kwargs={"required": False, "type": str, "help": "path to test instance set"},
)

InstanceSetTrainOptionalArgument = ArgumentContainer(
    names=["--instance-set-train"],
    kwargs={"required": False, "type": Path, "help": "path to training instance set"},
)

InstanceSetsReportArgument = ArgumentContainer(
    names=["--instance-sets", "--instance-set"],
    kwargs={
        "required": False,
        "nargs": "+",
        "type": str,
        "help": "Instance Set(s) to use for the report. If not "
        "specified, all Instance Sets are used.",
    },
)

InstancesPathArgument = ArgumentContainer(
    names=["instances_path"],
    kwargs={
        "metavar": "instances-path",
        "type": str,
        "help": "path to the instance set",
    },
)

InstancesPathRemoveArgument = ArgumentContainer(
    names=["instances_path"],
    kwargs={
        "metavar": "instances-path",
        "type": str,
        "help": "path to or nickname of the instance set",
    },
)

JobIDsArgument = ArgumentContainer(
    names=["--job-ids"],
    kwargs={
        "required": False,
        "nargs": "+",
        "type": str,
        "default": None,
        "help": "job ID(s) to use for the command",
    },
)

NicknameFeatureExtractorArgument = ArgumentContainer(
    names=["--nickname"],
    kwargs={"type": str, "help": "set a nickname for the feature extractor"},
)

NicknameInstanceSetArgument = ArgumentContainer(
    names=["--nickname"],
    kwargs={"type": str, "help": "set a nickname for the instance set"},
)

NicknamePortfolioArgument = ArgumentContainer(
    names=["--portfolio-name"],
    kwargs={
        "type": Path,
        "help": "Specify a name of the portfolio. "
        "If none is given, one will be generated.",
    },
)

NicknameSolverArgument = ArgumentContainer(
    names=["--nickname"], kwargs={"type": str, "help": "set a nickname for the solver"}
)

NoCopyArgument = ArgumentContainer(
    names=["--no-copy"],
    kwargs={
        "action": "store_true",
        "required": False,
        "help": "do not copy the source directory to "
        "the platform directory, but create a"
        " symbolic link instead",
    },
)

NoSavePlatformArgument = ArgumentContainer(
    names=["--no-save"],
    kwargs={
        "action": "store_false",
        "default": True,
        "required": False,
        "help": "do not save the platform upon re-initialisation.",
    },
)

RecomputeFeaturesArgument = ArgumentContainer(
    names=["--recompute"],
    kwargs={
        "action": "store_true",
        "help": "Re-run feature extractor for instances with "
        "previously computed features",
    },
)

RecomputePortfolioSelectorArgument = ArgumentContainer(
    names=["--recompute-portfolio-selector"],
    kwargs={
        "action": "store_true",
        "help": "force the construction of a new portfolio "
        "selector even when it already exists for the current "
        "feature and performance data.",
    },
)

RecomputeRunSolversArgument = ArgumentContainer(
    names=["--recompute"],
    kwargs={
        "action": "store_true",
        "help": "recompute the performance of all solvers on all instances",
    },
)

PerformanceDataJobsArgument = ArgumentContainer(
    names=["--performance-data-jobs"],
    kwargs={
        "action": "store_true",
        "help": "compute the remaining jobs in the Performance DataFrame",
    },
)

RebuildRunsolverArgument = ArgumentContainer(
    names=["--rebuild-runsolver"],
    kwargs={
        "action": "store_true",
        "required": False,
        "default": False,
        "help": "Clean the RunSolver executable and rebuild it.",
    },
)


SeedArgument = ArgumentContainer(
    names=["--seed"], kwargs={"type": int, "help": "seed to use for the command"}
)

SelectionScenarioArgument = ArgumentContainer(
    names=["--selection-scenario"],
    kwargs={"required": True, "type": Path, "help": "the selection scenario to use"},
)

SelectorAblationArgument = ArgumentContainer(
    names=["--solver-ablation"],
    kwargs={
        "required": False,
        "action": "store_true",
        "help": "construct a selector for each solver ablation combination",
    },
)

SettingsFileArgument = ArgumentContainer(
    names=["--settings-file"],
    kwargs={
        "type": Path,
        "help": "Specify the settings file to use in case you want"
        " supplement or override the default file.",
    },
)

SkipChecksArgument = ArgumentContainer(
    names=["--skip-checks"],
    kwargs={
        "dest": "run_checks",
        "default": True,
        "action": "store_false",
        "help": "Checks the solver's functionality by testing it on an instance "
        "and the pcs file, when applicable.",
    },
)

SnapshotArgument = ArgumentContainer(
    names=["snapshot_file_path"],
    kwargs={
        "metavar": "snapshot-file-path",
        "type": str,
        "help": "path to the snapshot file",
    },
)

SnapshotNameArgument = ArgumentContainer(
    names=["--name"],
    kwargs={"required": False, "type": str, "help": "name of the snapshot"},
)

SolverArgument = ArgumentContainer(
    names=["--solver"],
    kwargs={"required": True, "type": Path, "help": "path to solver"},
)

SolversArgument = ArgumentContainer(
    names=["--solvers", "--solver-paths", "--solver", "--solver-path"],
    kwargs={
        "required": False,
        "nargs": "+",
        "type": str,
        "help": "Specify the list of solvers to be "
        "used. If not specifed, all solvers "
        "known in Sparkle will be used.",
    },
)
SolverRemoveArgument = ArgumentContainer(
    names=["solver"],
    kwargs={
        "metavar": "solver",
        "type": str,
        "help": "name, path to or nickname of the solver",
    },
)

SolverPathArgument = ArgumentContainer(
    names=["solver_path"],
    kwargs={"metavar": "solver-path", "type": str, "help": "path to the solver"},
)

SolversReportArgument = ArgumentContainer(
    names=["--solvers", "--solver"],
    kwargs={
        "nargs": "+",
        "type": str,
        "default": None,
        "help": "Solver(s) to use for the report. If not specified, all solvers are "
        "included.",
    },
)

TestSetRunAllConfigurationArgument = ArgumentContainer(
    names=["--test-set-run-all-configurations"],
    kwargs={
        "required": False,
        "action": "store_true",
        "help": "run all found configurations on the test set",
    },
)

UseFeaturesArgument = ArgumentContainer(
    names=["--use-features"],
    kwargs={
        "required": False,
        "action": "store_true",
        "help": "use the training set's features for configuration",
    },
)

SolutionVerifierArgument = ArgumentContainer(
    names=["--solution-verifier"],
    kwargs={
        "type": str,
        "default": None,
        "help": "the class name of the solution verifier to use "
        "for the Solver. If it is a Path, will resolve as "
        "a SolutionFileVerifier class with the specified "
        "Path instead.",
    },
)

ObjectiveArgument = ArgumentContainer(
    names=["--objective"], kwargs={"type": str, "help": "the objective to use."}
)
